import boto3


class CodeBuildUtils:
    """ Class for interacting with AWS CodeBuild (for Docker builds) """

    def __init__(self, client=None):
        if client:
            self.client = client
        else:
            self.client = boto3.client('codebuild')

    def list_projects(self) -> [str]:
        """ Returns a list of projects on the current AWS Account CodeBuild"""
        return self.client.list_projects().get('projects', [])

    def run_project_build(self, *, project_name: str) -> dict:
        """ Runs the given project build. """
        return self.client.start_build(projectName=project_name)

    def run_project_build_with_overrides(self, *, project_name, branch, env_overrides):
        """ Runs the given project build with branch and environment variable overrides """
        return self.client.start_build(
            projectName=project_name,
            sourceVersion=branch,
            environmentVariablesOverride=[
                {
                    'name': k,
                    'value': v,
                    'type': 'PLAINTEXT'
                } for k, v in env_overrides.items()
            ]
        )



# ------------------------------------------------------------------------------------------------------
# Command-line utility to update (post, patch, upsert) portal objects for SMaHT/CGAP/Fourfront.
# ------------------------------------------------------------------------------------------------------
# Example commands:
# update-portal-object --post file_format.json
# update-portal-object --upsert directory-with-schema-named-dot-json-files
# update-portal-object --patch file-not-named-for-schema-name.json --schema UnalignedReads
# --------------------------------------------------------------------------------------------------

import argparse
from functools import lru_cache
import glob
import io
import json
import os
import sys
from typing import Callable, List, Optional, Tuple, Union
from dcicutils.command_utils import yes_or_no
from dcicutils.common import ORCHESTRATED_APPS, APP_SMAHT
from dcicutils.ff_utils import delete_metadata, purge_metadata
from dcicutils.misc_utils import get_error_message, PRINT
from dcicutils.portal_utils import Portal as PortalFromUtils


class Portal(PortalFromUtils):

    def delete_metadata(self, object_id: str) -> Optional[dict]:
        if isinstance(object_id, str) and object_id and self.key:
            return delete_metadata(obj_id=object_id, key=self.key)
        return None

    def purge_metadata(self, object_id: str) -> Optional[dict]:
        if isinstance(object_id, str) and object_id and self.key:
            return purge_metadata(obj_id=object_id, key=self.key)
        return None


_DEFAULT_APP = "smaht"
_SMAHT_ENV_ENVIRON_NAME = "SMAHT_ENV"

# Schema properties to ignore (by default) for the view schema usage.
_SCHEMAS_IGNORE_PROPERTIES = [
    "date_created",
    "last_modified",
    "principals_allowed",
    "submitted_by",
    "schema_version"
]

_SCHEMA_ORDER = [  # See: smaht-portal/src/encoded/project/loadxl.py
    "access_key",
    "user",
    "consortium",
    "submission_center",
    "file_format",
    "quality_metric",
    "output_file",
    "reference_file",
    "reference_genome",
    "software",
    "tracking_item",
    "workflow",
    "workflow_run",
    "meta_workflow",
    "meta_workflow_run",
    "image",
    "document",
    "static_section",
    "page",
    "filter_set",
    "higlass_view_config",
    "ingestion_submission",
    "ontology_term",
    "protocol",
    "donor",
    "demographic",
    "medical_history",
    "diagnosis",
    "exposure",
    "family_history",
    "medical_treatment",
    "death_circumstances",
    "tissue_collection",
    "tissue",
    "histology",
    "cell_line",
    "cell_culture",
    "cell_culture_mixture",
    "preparation_kit",
    "treatment",
    "sample_preparation",
    "tissue_sample",
    "cell_culture_sample",
    "cell_sample",
    "analyte",
    "analyte_preparation",
    "assay",
    "library",
    "library_preparation",
    "sequencer",
    "basecalling",
    "sequencing",
    "file_set",
    "unaligned_reads",
    "aligned_reads",
    "variant_calls",
]


def main():

    parser = argparse.ArgumentParser(description="View Portal object.")
    parser.add_argument("--env", "-e", type=str, required=False, default=None,
                        help=f"Environment name (key from ~/.smaht-keys.json).")
    parser.add_argument("--app", type=str, required=False, default=None,
                        help=f"Application name (one of: smaht, cgap, fourfront).")
    parser.add_argument("--schema", type=str, required=False, default=None,
                        help="Use named schema rather than infer from post/patch/upsert file name.")
    parser.add_argument("--post", type=str, required=False, default=None, help="POST data.")
    parser.add_argument("--patch", type=str, required=False, default=None, help="PATCH data.")
    parser.add_argument("--upsert", type=str, required=False, default=None, help="Upsert data.")
    parser.add_argument("--delete", type=str, required=False, default=None, help="Delete data.")
    parser.add_argument("--purge", type=str, required=False, default=None, help="Purge data.")
    parser.add_argument("--confirm", action="store_true", required=False, default=False, help="Confirm before action.")
    parser.add_argument("--verbose", action="store_true", required=False, default=False, help="Verbose output.")
    parser.add_argument("--quiet", action="store_true", required=False, default=False, help="Quiet output.")
    parser.add_argument("--debug", action="store_true", required=False, default=False, help="Debugging output.")
    args = parser.parse_args()

    def usage(message: Optional[str] = None) -> None:
        nonlocal parser
        _print(message) if isinstance(message, str) else None
        parser.print_help()
        sys.exit(1)

    if app := args.app:
        if (app not in ORCHESTRATED_APPS) and ((app := app.lower()) not in ORCHESTRATED_APPS):
            usage(f"ERROR: Unknown app name; must be one of: {' | '.join(ORCHESTRATED_APPS)}")
    else:
        app = APP_SMAHT

    portal = _create_portal(env=args.env, app=app, verbose=args.verbose, debug=args.debug)

    if explicit_schema_name := args.schema:
        schema, explicit_schema_name = _get_schema(portal, explicit_schema_name)
        if not schema:
            usage(f"ERROR: Unknown schema name: {args.schema}")

    if not (args.post or args.patch or args.upsert or args.delete or args.purge):
        usage()

    if args.post:
        _post_or_patch_or_upsert(portal=portal,
                                 file_or_directory=args.post,
                                 explicit_schema_name=explicit_schema_name,
                                 update_function=post_data,
                                 update_action_name="POST",
                                 confirm=args.confirm, verbose=args.verbose, quiet=args.quiet, debug=args.debug)
    if args.patch:
        _post_or_patch_or_upsert(portal=portal,
                                 file_or_directory=args.patch,
                                 explicit_schema_name=explicit_schema_name,
                                 update_function=patch_data,
                                 update_action_name="PATCH",
                                 confirm=args.confirm, verbose=args.verbose, quiet=args.quiet, debug=args.debug)
    if args.upsert:
        _post_or_patch_or_upsert(portal=portal,
                                 file_or_directory=args.upsert,
                                 explicit_schema_name=explicit_schema_name,
                                 update_function=upsert_data,
                                 update_action_name="UPSERT",
                                 confirm=args.confirm, verbose=args.verbose, quiet=args.quiet, debug=args.debug)

    if args.delete:
        if not portal.get_metadata(args.delete, raise_exception=False):
            _print(f"Cannot find given object: {args.delete}")
            sys.exit(1)
        if yes_or_no(f"Do you really want to delete this item: {args.delete} ?"):
            portal.delete_metadata(args.delete)

    if args.purge:
        if not portal.get_metadata(args.purge, raise_exception=False):
            _print(f"Cannot find given object: {args.purge}")
            sys.exit(1)
        if yes_or_no(f"Do you really want to purge this item: {args.purge} ?"):
            portal.delete_metadata(args.purge)
            portal.purge_metadata(args.purge)


def _post_or_patch_or_upsert(portal: Portal, file_or_directory: str,
                             explicit_schema_name: str,
                             update_function: Callable, update_action_name: str,
                             confirm: bool = False, verbose: bool = False,
                             quiet: bool = False, debug: bool = False) -> None:

    def is_schema_name_list(portal: Portal, keys: list) -> bool:
        if isinstance(keys, list):
            for key in keys:
                if portal.get_schema(key) is None:
                    return False
                return True
        return False

    def post_or_patch_or_upsert(portal: Portal, file: str, schema_name: Optional[str],
                                confirm: bool = False, verbose: bool = False,
                                quiet: bool = False, debug: bool = False) -> None:

        nonlocal update_function, update_action_name
        if not quiet:
            _print(f"Processing {update_action_name} file: {file}")
        if data := _read_json_from_file(file):
            if isinstance(data, dict):
                if isinstance(schema_name, str) and schema_name:
                    if debug:
                        _print(f"DEBUG: File ({file}) contains an object of type: {schema_name}")
                    update_function(portal, data, schema_name, confirm=confirm,
                                    file=file, verbose=verbose, debug=debug)
                elif is_schema_name_list(portal, list(data.keys())):
                    if debug:
                        _print(f"DEBUG: File ({file}) contains a dictionary of schema names.")
                    for schema_name in data:
                        if isinstance(schema_data := data[schema_name], list):
                            if debug:
                                _print(f"DEBUG: Processing {update_action_name}s for type: {schema_name}")
                            for index, item in enumerate(schema_data):
                                update_function(portal, item, schema_name, confirm=confirm,
                                                file=file, index=index, verbose=verbose, debug=debug)
                        else:
                            _print(f"WARNING: File ({file}) contains schema item which is not a list: {schema_name}")
                else:
                    _print(f"WARNING: File ({file}) contains unknown item type.")
            elif isinstance(data, list):
                if debug:
                    _print(f"DEBUG: File ({file}) contains a list of objects of type: {schema_name}")
                for index, item in enumerate(data):
                    update_function(portal, item, schema_name, confirm=confirm,
                                    file=file, index=index, verbose=verbose, debug=debug)
            if debug:
                _print(f"DEBUG: Processing {update_action_name} file done: {file}")

    if os.path.isdir(file_or_directory):
        if ((files := glob.glob(os.path.join(file_or_directory, "*.json"))) and
            (files_and_schemas := _file_names_to_ordered_file_and_schema_names(portal, files))):  # noqa
            for file_and_schema in files_and_schemas:
                if not (file := file_and_schema[0]):
                    continue
                if not (schema_name := file_and_schema[1]) and not (schema_name := explicit_schema_name):
                    _print(f"ERROR: Schema cannot be inferred from file name and --schema not specified: {file}")
                    continue
                post_or_patch_or_upsert(portal, file_and_schema[0], schema_name=schema_name,
                                        confirm=confirm, quiet=quiet, verbose=verbose, debug=debug)
    elif os.path.isfile(file := file_or_directory):
        if ((schema_name := _get_schema_name_from_schema_named_json_file_name(portal, file)) or
            (schema_name := explicit_schema_name)):  # noqa
            post_or_patch_or_upsert(portal, file, schema_name=schema_name,
                                    confirm=confirm, quiet=quiet, verbose=verbose, debug=debug)
        else:
            post_or_patch_or_upsert(portal, file, schema_name=schema_name,
                                    confirm=confirm, quiet=quiet, verbose=verbose, debug=debug)
            # _print(f"ERROR: Schema cannot be inferred from file name and --schema not specified: {file}")
            # return
    else:
        _print(f"ERROR: Cannot find file or directory: {file_or_directory}")


def post_data(portal: Portal, data: dict, schema_name: str, confirm: bool = False,
              file: Optional[str] = None, index: int = 0,
              verbose: bool = False, debug: bool = False) -> None:
    if not (identifying_path := portal.get_identifying_path(data, portal_type=schema_name)):
        if isinstance(file, str) and isinstance(index, int):
            _print(f"ERROR: Item for POST has no identifying property: {file} (#{index + 1})")
        else:
            _print(f"ERROR: Item for POST has no identifying property.")
        return
    if portal.get_metadata(identifying_path, raise_exception=False):
        _print(f"ERROR: Item for POST already exists: {identifying_path}")
        return
    if (confirm is True) and not yes_or_no(f"POST data for: {identifying_path} ?"):
        return
    if verbose:
        _print(f"POST {schema_name} item: {identifying_path}")
    try:
        portal.post_metadata(schema_name, data)
        if debug:
            _print(f"DEBUG: POST {schema_name} item done: {identifying_path}")
    except Exception as e:
        _print(f"ERROR: Cannot POST {schema_name} item: {identifying_path}")
        _print(get_error_message(e))
        return


def patch_data(portal: Portal, data: dict, schema_name: str, confirm: bool = False,
               file: Optional[str] = None, index: int = 0,
               verbose: bool = False, debug: bool = False) -> None:
    if not (identifying_path := portal.get_identifying_path(data, portal_type=schema_name)):
        if isinstance(file, str) and isinstance(index, int):
            _print(f"ERROR: Item for PATCH has no identifying property: {file} (#{index + 1})")
        else:
            _print(f"ERROR: Item for PATCH has no identifying property.")
        return
    if not portal.get_metadata(identifying_path, raise_exception=False):
        _print(f"ERROR: Item for PATCH does not already exist: {identifying_path}")
        return
    if (confirm is True) and not yes_or_no(f"PATCH data for: {identifying_path}"):
        return
    if verbose:
        _print(f"PATCH {schema_name} item: {identifying_path}")
    try:
        portal.patch_metadata(identifying_path, data)
        if debug:
            _print(f"DEBUG: PATCH {schema_name} item OK: {identifying_path}")
    except Exception as e:
        _print(f"ERROR: Cannot PATCH {schema_name} item: {identifying_path}")
        _print(e)
        return


def upsert_data(portal: Portal, data: dict, schema_name: str, confirm: bool = False,
                file: Optional[str] = None, index: int = 0,
                verbose: bool = False, debug: bool = False) -> None:
    if not (identifying_path := portal.get_identifying_path(data, portal_type=schema_name)):
        if isinstance(file, str) and isinstance(index, int):
            _print(f"ERROR: Item for UPSERT has no identifying property: {file} (#{index + 1})")
        else:
            _print(f"ERROR: Item for UPSERT has no identifying property.")
        return
    exists = portal.get_metadata(identifying_path, raise_exception=False)
    if ((confirm is True) and not yes_or_no(f"{'PATCH' if exists else 'POST'} data for: {identifying_path} ?")):
        return
    if verbose:
        _print(f"{'PATCH' if exists else 'POST'} {schema_name} item: {identifying_path}")
    try:
        portal.post_metadata(schema_name, data) if not exists else portal.patch_metadata(identifying_path, data)
        if debug:
            _print(f"DEBUG: UPSERT {schema_name} item OK: {identifying_path}")
    except Exception as e:
        _print(f"ERROR: Cannot UPSERT {schema_name} item: {identifying_path}")
        _print(e)
        return


def _create_portal(env: Optional[str] = None, app: Optional[str] = None,
                   verbose: bool = False, debug: bool = False) -> Optional[Portal]:

    env_from_environ = None
    if not env and (app == APP_SMAHT):
        if env := os.environ.get(_SMAHT_ENV_ENVIRON_NAME):
            env_from_environ = True
    if not (portal := Portal(env, app=app) if env or app else None):
        return None
    if verbose:
        if (env := portal.env) or (env := os.environ(_SMAHT_ENV_ENVIRON_NAME)):
            _print(f"Portal environment"
                   f"{f' (from {_SMAHT_ENV_ENVIRON_NAME})' if env_from_environ else ''}: {portal.env}")
        if portal.keys_file:
            _print(f"Portal keys file: {portal.keys_file}")
        if portal.key_id:
            _print(f"Portal key prefix: {portal.key_id[0:2]}******")
        if portal.server:
            _print(f"Portal server: {portal.server}")
    return portal


def _read_json_from_file(file: str) -> Optional[dict]:
    try:
        if not os.path.exists(file):
            return None
        with io.open(file, "r") as f:
            try:
                return json.load(f)
            except Exception:
                _print(f"ERROR: Cannot load JSON from file: {file}")
                return None
    except Exception:
        _print(f"ERROR: Cannot open file: {file}")
        return None


def _file_names_to_ordered_file_and_schema_names(portal: Portal,
                                                 files: Union[List[str], str]) -> List[Tuple[str, Optional[str]]]:
    results = []
    if isinstance(files, str):
        files = [files]
    if not isinstance(files, list):
        return results
    for file in files:
        if isinstance(file, str) and file:
            results.append((file, _get_schema_name_from_schema_named_json_file_name(portal, file)))
    ordered_results = []
    for schema_name in _SCHEMA_ORDER:
        schema_name = portal.schema_name(schema_name)
        if result := next((item for item in results if item[1] == schema_name), None):
            ordered_results.append(result)
            results.remove(result)
    ordered_results.extend(results) if results else None
    return ordered_results


def _get_schema_name_from_schema_named_json_file_name(portal: Portal, value: str) -> Optional[str]:
    try:
        if not value.endswith(".json"):
            return None
        _, schema_name = _get_schema(portal, os.path.basename(value[:-5]))
        return schema_name
    except Exception:
        return False


@lru_cache(maxsize=1)
def _get_schemas(portal: Portal) -> Optional[dict]:
    return portal.get_schemas()


@lru_cache(maxsize=100)
def _get_schema(portal: Portal, name: str) -> Tuple[Optional[dict], Optional[str]]:
    if portal and name and (name := name.replace("_", "").replace("-", "").strip().lower()):
        if schemas := _get_schemas(portal):
            for schema_name in schemas:
                if schema_name.replace("_", "").replace("-", "").strip().lower() == name.lower():
                    return schemas[schema_name], schema_name
    return None, None


def _print(*args, **kwargs) -> None:
    PRINT(*args, **kwargs)
    sys.stdout.flush()


if __name__ == "__main__":
    main()



import copy

from typing import Any, Dict, List, Optional, Tuple, Union
from .common import AnyJsonData
from .env_utils import EnvUtils, public_env_name
from .ff_utils import get_metadata
from .misc_utils import AbstractVirtualApp, ignored, ignorable, PRINT, remove_empty_properties, to_camel_case
from .sheet_utils import (
    LoadTableError, prefer_number, TabbedJsonSchemas,
    Header, Headers, TabbedHeaders, ParsedHeader, ParsedHeaders, TabbedParsedHeaders, SheetCellValue, TabbedSheetData,
    TableSetManagerRegistry, AbstractTableSetManager, InsertsManager, TableSetManager, load_table_set,
)
from .validation_utils import SchemaManager, validate_data_against_schemas


PatchPrototype = Dict
TabbedPatchPrototypes = Dict[str, PatchPrototype]
ARRAY_VALUE_DELIMITER = "|"


class TypeHintContext:

    @classmethod
    def schema_exists(cls, schema_name: str) -> bool:  # noQA - PyCharm complains wrongly about return value
        ignored(schema_name)
        raise NotImplementedError(f"{cls.__name__}.schema_exists(...) is not implemented.")

    @classmethod
    def validate_ref(cls, item_type: str, item_ref: str) -> str:  # noQA - PyCharm complains wrongly about return value
        ignored(item_type, item_ref)
        raise NotImplementedError(f"{cls.__name__}.validate_ref(...) is not implemented.")

    @classmethod
    def note_problem(cls, problem: str):
        ignored(problem)
        raise NotImplementedError(f"{cls.__name__}.note_problem(...) is not implemented.")

    def __str__(self):
        return f"<{self.__class__.__name__} {id(self)}>"


class ValidationProblem(Exception):
    def __init__(self, problems: Optional[dict] = None):
        self.problems = problems


class TypeHint:
    def __init__(self):
        self.is_array = False

    def apply_hint(self, value, src):
        return value

    def __str__(self):
        return f"<{self.__class__.__name__}>"

    def __repr__(self):
        return self.__str__()


class BoolHint(TypeHint):
    def __init__(self):
        super().__init__()

    # We could use other ways to do this, such as initial substring, but this is more likely to be right.
    # Then again, we might want to consder accepting athers like 'yes/no', 'y/n', 'on/off', '1/0'.
    TRUE_VALUES = ['true', 't']
    FALSE_VALUES = ['false', 'f']

    def apply_hint(self, value, src):
        if isinstance(value, str) and value:
            l_value = value.lower()
            if l_value in self.TRUE_VALUES:
                return True
            elif l_value in self.FALSE_VALUES:
                return False
        return super().apply_hint(value, src)


class NumHint(TypeHint):

    PREFERENCE_MAP = {'number': 'number', 'integer': 'integer'}

    def __init__(self, declared_type: Optional[str] = None):
        if declared_type is None:
            declared_type = 'number'
        self.preferred_type = self.PREFERENCE_MAP.get(declared_type)
        super().__init__()

    def apply_hint(self, value, src):
        if isinstance(value, str) and value:
            if self.preferred_type:
                return prefer_number(value, kind=self.preferred_type)
            else:
                return value
        return super().apply_hint(value, src)


class EnumHint(TypeHint):

    def __str__(self):
        return f"<EnumHint {','.join(f'{key}={val}' for key, val in self.value_map.items())}>"

    def __init__(self, value_map):
        self.value_map = value_map
        super().__init__()

    def apply_hint(self, value, src):
        if isinstance(value, str):
            if value in self.value_map:
                result = self.value_map[value]
                return result
            else:
                lvalue = value.lower()
                found = []
                for lkey, key in self.value_map.items():
                    if lkey.startswith(lvalue):
                        found.append(lkey)
                if len(found) == 1:
                    [only_found] = found
                    result = self.value_map[only_found]
                    return result
        return super().apply_hint(value, src)


class ArrayHint(TypeHint):
    def __init__(self):
        super().__init__()

    def apply_hint(self, value, src):
        if value is None or value == '':
            return []
        if isinstance(value, str):
            if not value:
                return []
            return [value.strip() for value in value.split(ARRAY_VALUE_DELIMITER)]
        return super().apply_hint(value, src)


class StringHint(TypeHint):
    def __init__(self):
        super().__init__()

    def apply_hint(self, value, src):
        return str(value).strip() if value is not None else ""


class RefHint(TypeHint):

    def __str__(self):
        return f"<RefHint {self.schema_name} context={self.context}>"

    def __init__(self, schema_name: str, required: bool, context: TypeHintContext):
        self.schema_name = schema_name
        self.context = context
        self.required = required
        super().__init__()

    def apply_hint(self, value, src):
        if self.is_array and isinstance(value, str):
            value = [value.strip() for value in value.split(ARRAY_VALUE_DELIMITER)] if value else []
        if self.is_array and isinstance(value, list):
            for item in value:
                self._apply_ref_hint(item, src)
        else:
            self._apply_ref_hint(value, src)
        return value

    def _apply_ref_hint(self, value, src):
        if not value and self.required:
            raise ValidationProblem(f"No required reference (linkTo) value for: {self.schema_name}"
                                    f"{f' from {src}' if src else ''}")
        if value and not self.context.validate_ref(item_type=self.schema_name, item_ref=value):
            raise ValidationProblem(f"Cannot resolve reference (linkTo) for: {self.schema_name}"
                                    f"{f'/{value}' if value else ''}{f' from {src}' if src else ''}")
        return value


class OptionalTypeHints:

    def __init__(self, positional_hints: Optional[List[Optional[TypeHint]]] = None,
                 positional_breadcrumbs: Optional[List[Union[List, Tuple]]] = None):
        self.other_hints: Dict[Any, TypeHint] = {}
        self.positional_hints: List[Optional[TypeHint]] = [] if positional_hints is None else positional_hints
        if positional_breadcrumbs and positional_hints:
            n = len(positional_breadcrumbs)
            if n != len(positional_hints):
                raise Exception("positional_hints and positional_breadcrumbs must have the same length.")
            for i in range(n):
                # for convenience, we accept this as a list or tuple, but it must be a tuple to be a key
                breadcrumbs = tuple(positional_breadcrumbs[i])
                if not isinstance(breadcrumbs, tuple):
                    raise Exception(f"Each of the positional breadcrumbs must be a tuple: {breadcrumbs}")
                hint = positional_hints[i]
                self.other_hints[breadcrumbs] = hint

    def __getitem__(self, key: Any) -> Optional[TypeHint]:
        """
        For enumerated positional information, we consult our initial type vector.
        For other situations, we do a general lookup of the hint in our lookup table.
        """
        if isinstance(key, int):
            hints = self.positional_hints
            if key < 0:
                raise ValueError(f"Negative hint positions are not allowed: {key}")
            elif key >= len(hints):
                return None
            else:
                return hints[key]
        elif isinstance(key, tuple):  # a parsed header (or schema breadcrumbs)
            return self.other_hints.get(key)
        else:
            raise ValueError(f"Key of unexpected type for OptionalTypeHints: {key}")

    def __setitem__(self, key: Any, value: TypeHint):
        if isinstance(key, int):
            raise ValueError(f"Cannot assign OptionalTypeHints by position after initial creation: {key!r}")
        elif key in self.other_hints:
            raise ValueError(f"Attempt to redefine OptionalTypeHint key {key!r}.")
        elif isinstance(key, tuple):
            self.other_hints[key] = value
        else:
            raise ValueError(f"Attempt to set an OptionalTypeHints key to other than a breadcrumbs tuple: {key!r}")


class AbstractStructureManager(AbstractTableSetManager):

    pass


class ItemTools:
    """
    Implements operations on table-related data without pre-supposing the specific representation of the table.
    It is assumed this can be used for data that was obtained from .json, .csv, .tsv, and .xlsx files because
    it does not presuppose the source of the data nor where it will be written to.

    For the purpose of this class:

    * a 'header' is a string representing the top of a column.

    * a 'parsed header' is a list of strings and/or ints, after splitting at uses of '#' or '.', so that
      "a.b.c" is represented as ["a", "b", "c"], and "x.y#0" is represented as ["x", "y", 0], and representing
      each numeric token as an int instead of a string.

    * a 'headers' object is just a list of strings, each of which is a 'header'.

    * a 'parsed headers' object is a non-empty list of lists, each of which is a 'parsed header'.
      e..g., the headers ["a.b.c", "x.y#0"] is represented as parsed hearders [["a", "b", "c"], ["x", "y", 0]].

   """

    @classmethod
    def parse_sheet_header(cls, header: Header) -> ParsedHeader:
        result = []
        token = ""
        for i in range(len(header)):
            ch = header[i]
            if ch == '.' or ch == '#':
                if token:
                    result.append(int(token) if token.isdigit() else token)
                    token = ""
            else:
                token += ch
        if token:
            result.append(int(token) if token.isdigit() else token)
        return result

    @classmethod
    def parse_sheet_headers(cls, headers: Headers) -> ParsedHeaders:
        return [cls.parse_sheet_header(header)
                for header in headers]

    @classmethod
    def compute_patch_prototype(cls, parsed_headers: ParsedHeaders):
        prototype = {}
        for parsed_header in parsed_headers:
            parsed_header0 = parsed_header[0] if parsed_header else ""
            if isinstance(parsed_header0, int):
                raise LoadTableError(f"A header cannot begin with a numeric ref: {parsed_header0}")
            cls.assure_patch_prototype_shape(parent=prototype, keys=parsed_header)
        return prototype

    @classmethod
    def assure_patch_prototype_shape(cls, *, parent: Union[Dict, List], keys: ParsedHeader):
        key0 = None
        more_keys = None
        [key0, *more_keys] = keys if keys else [None]
        key1 = more_keys[0] if more_keys else None
        if isinstance(key1, int):
            placeholder = []
        elif isinstance(key1, str):
            placeholder = {}
        else:
            placeholder = None
        if isinstance(key0, int):
            n = len(parent)
            if key0 == n:
                parent.append(placeholder)
            elif key0 > n:
                raise LoadTableError("Numeric items must occur sequentially.")
        elif isinstance(key0, str):
            if key0 not in parent:
                parent[key0] = placeholder
        if key1 is not None:
            cls.assure_patch_prototype_shape(parent=parent[key0], keys=more_keys)
        return parent

    INSTAGUIDS_ENABLED = False  # Experimental feature not enabled by default

    @classmethod
    def parse_item_value(cls, value: SheetCellValue,
                         apply_heuristics: bool = False, split_pipe: bool = False) -> AnyJsonData:
        """
        Returns the item value unmodified, unless apply_heuristics=True is given,
        in which case heuristics ARE applied. This is intended to be used for spreadsheet
        values that look like non-strings and should perhaps be interepreted as such.

        This is a vestige of an older plan to have these things happen magically behind the scenes early in
        the process. Unfortunately, that was found to impede correct processing later, so now this is disabled
        by default. It may still be useful in some cases when dealing with data that has no schema, so the
        functionality is still here and must be explicitly requested.

        :param value: a value in a table (such as a spreadsheet)
        :param apply_heuristics: whether to apply heuristic coercions based on what the value looks like (default False)
        :param split_pipe: whether to apply the 'split pipe' heuristic, changing 'a|1' to ['a', 1], even if
           apply_heuristics=True was given (default False)
        """
        if not apply_heuristics:
            # In order to not interfere with schema-driven processing, we mostly default to
            # NOT applying heuristics. You have to ask for them explicitly if you want them.
            # -kmp 23-Oct-2023
            return value
        if isinstance(value, str):
            lvalue = value.lower()
            if lvalue == 'true':
                return True
            elif lvalue == 'false':
                return False
            # elif lvalue == 'null' or lvalue == '':
            elif lvalue == 'null':
                return None
            elif lvalue == '':
                return lvalue
            elif split_pipe and '|' in value:
                if value == '|':  # Use '|' for []
                    return []
                else:
                    if value.endswith("|"):  # Use 'foo|' for ['foo']
                        value = value[:-1]
                    return [cls.parse_item_value(subvalue, apply_heuristics=apply_heuristics, split_pipe=split_pipe)
                            for subvalue in value.split('|')]
            else:
                # Doug points out that the schema might not agree, might want a string representation of a number.
                # At this semantic layer, this might be a bad choice.
                return prefer_number(value)
        else:  # presumably a number (int or float)
            return value

    @classmethod
    def set_path_value(cls, datum: Union[List, Dict], path: ParsedHeader, value: Any, force: bool = False):
        # if (value is None or value == '') and not force:
        if value is None and not force:
            return
        [key, *more_path] = path if path else [None]
        if not more_path:
            datum[key] = value
        else:
            cls.set_path_value(datum[key], more_path, value)

    @classmethod
    def find_type_hint_for_subschema(cls, subschema: Any, required: bool = False,
                                     context: Optional[TypeHintContext] = None):
        if subschema is not None:
            t = subschema.get('type')
            if t == 'string':
                enum = subschema.get('enum')
                if enum:
                    mapping = {e.lower(): e for e in enum}
                    return EnumHint(mapping)
                link_to = subschema.get('linkTo')
                if link_to and context.schema_exists(link_to):
                    return RefHint(schema_name=link_to, required=required, context=context)
                return StringHint()
            elif t in ('integer', 'number'):
                return NumHint(declared_type=t)
            elif t == 'boolean':
                return BoolHint()
            elif t == 'array':
                array_type_hint = cls.find_type_hint_for_subschema(subschema.get("items"),
                                                                   required=required, context=context)
                if type(array_type_hint) is RefHint:
                    array_type_hint.is_array = True
                    return array_type_hint
                return ArrayHint()

    @classmethod
    def find_type_hint_for_parsed_header(cls, parsed_header: Optional[ParsedHeader], schema: Any,
                                         context: Optional[TypeHintContext] = None):
        def finder(subheader, subschema):
            if not parsed_header:
                return None
            else:
                [key1, *other_headers] = subheader
                if isinstance(key1, str) and isinstance(subschema, dict):
                    if subschema.get('type') == 'object':
                        subsubschema = subschema.get('properties', {}).get(key1)
                        if not other_headers:
                            required = key1 and subschema and key1 in subschema.get('required', [])
                            hint = cls.find_type_hint_for_subschema(subsubschema, required=required, context=context)
                            if hint:
                                return hint
                            else:
                                pass  # fall through to asking super()
                        else:
                            return finder(subheader=other_headers, subschema=subsubschema)

        return finder(subheader=parsed_header, subschema=schema)


ITEM_MANAGER_REGISTRY = TableSetManagerRegistry()


class InflatableTabbedDataManager:
    """
    This tool can be used independently of the item tools. It doesn't involve schemas, but it does allow the
    inflation of a table with dotted names to structures. e.g., a table with headers mother.name, mother.age,
    father.name, and father.age, as in
      data = load_table_set(<some-file>)
    to bring in the flat representation with:
      {"mother.name": <mother.name>, "mother.age": <mother.age>, ...}
    one can use inflate(data) to get:
      {"mother": {"name": <mother.name>, "age": <mother.age>},
       "father:  {"name": <father.name>, "age": <father.age>}}
    Note, too, that although data != inflate(data), once inflated, inflate(inflate(data)) == inflate(data).
    """

    def __init__(self, tabbed_sheet_data: TabbedSheetData, apply_heuristics: bool = False):
        self.tabbed_sheet_data: TabbedSheetData = tabbed_sheet_data
        self.apply_heuristics = apply_heuristics
        self.headers_by_tab_name: TabbedHeaders = InsertsManager.extract_tabbed_headers(tabbed_sheet_data)
        self.parsed_headers_by_tab_name: TabbedParsedHeaders = {
            tab_name: ItemTools.parse_sheet_headers(headers)
            for tab_name, headers in self.headers_by_tab_name.items()
        }
        self.patch_prototypes_by_tab_name: TabbedPatchPrototypes = {
            tab_name: ItemTools.compute_patch_prototype(parsed_headers)
            for tab_name, parsed_headers in self.parsed_headers_by_tab_name.items()
        }

    @property
    def tab_names(self):
        return list(self.tabbed_sheet_data.keys())

    def inflate_tabs(self):
        return {tab_name: self.inflate_tab(tab_name)
                for tab_name in self.tab_names}

    def inflate_tab(self, tab_name: str):
        prototype = self.patch_prototypes_by_tab_name[tab_name]
        parsed_headers = self.parsed_headers_by_tab_name[tab_name]
        result = [self.inflate_row(row, prototype=prototype, parsed_headers=parsed_headers)
                  for row in self.tabbed_sheet_data[tab_name]]
        return result

    def inflate_row(self, row: Dict, *, prototype: Dict, parsed_headers: ParsedHeaders):
        patch_item = copy.deepcopy(prototype)
        for column_number, column_value in enumerate(row.values()):
            parsed_value = ItemTools.parse_item_value(column_value, apply_heuristics=self.apply_heuristics)
            ItemTools.set_path_value(patch_item, parsed_headers[column_number], parsed_value)
        return patch_item

    @classmethod
    def inflate(cls, tabbed_sheet_data: TabbedSheetData, apply_heuristics: bool = False):
        inflater = cls(tabbed_sheet_data, apply_heuristics=apply_heuristics)
        inflated = inflater.inflate_tabs()
        return inflated


inflate = InflatableTabbedDataManager.inflate


def load_table_structures(filename: str, *, apply_heuristics: bool = True,
                          tab_name: Optional[str] = None, escaping: Optional[bool] = None, **kwargs):
    """This differs from load_table_set only in that it inflates the content. It does not apply schemas."""
    tabbed_rows = load_table_set(filename=filename, tab_name=tab_name, escaping=escaping, **kwargs)
    tabbed_structures = inflate(tabbed_rows, apply_heuristics=apply_heuristics)
    return tabbed_structures


class TableChecker(InflatableTabbedDataManager, TypeHintContext):

    def __init__(self, tabbed_sheet_data: TabbedSheetData, *, flattened: bool,
                 override_schemas: Optional[TabbedJsonSchemas] = None,
                 portal_env: Optional[str] = None, portal_vapp: Optional[AbstractVirtualApp] = None,
                 apply_heuristics: bool = False):

        self.flattened = flattened
        # if not flattened:
        #     # TODO: Need to implement something that depends on this flattened attribute.
        #     # Also, it's possible that we can default this once we see if the new strategy is general-purpose,
        #     # rather than it being a required argument. But for now let's require it be passed.
        #     # -kmp 25-Oct-2023
        #     raise ValueError("Only flattened=True is supported by TableChecker for now.")

        if portal_env is None and portal_vapp is None:
            portal_env = public_env_name(EnvUtils.PRD_ENV_NAME)
        # InflatableTabbedDataManager supplies:
        #   self.tabbed_sheet_data: TabbedSheetData =
        #   self.headers_by_tab_name: TabbedHeaders =
        #   self.parsed_headers_by_tab_name: TabbedParsedHeaders =
        #   self.patch_prototypes_by_tab_name: TabbedPatchPrototypes =
        self._problems: List[str] = []
        super().__init__(tabbed_sheet_data=tabbed_sheet_data, apply_heuristics=apply_heuristics)
        self.portal_env = portal_env
        self.portal_vapp = portal_vapp
        self.schema_manager: SchemaManager = SchemaManager(portal_env=portal_env, portal_vapp=portal_vapp,
                                                           override_schemas=override_schemas)
        schema_names_to_fetch = [key for key, value in tabbed_sheet_data.items() if value]
        self.schemas = self.schema_manager.fetch_relevant_schemas(schema_names_to_fetch)
        self.lookup_tables_by_tab_name: Dict[str, Dict[str, Dict]] = {
            tab_name: self.build_lookup_table_for_tab(tab_name, rows=rows)
            for tab_name, rows in tabbed_sheet_data.items()
        }
        self.type_hints_by_tab_name: Dict[str, OptionalTypeHints] = {
            tab_name: self.compile_type_hints(tab_name)
            for tab_name in self.tab_names
        }

    def schema_for_tab(self, tab_name: str) -> dict:
        # Once our class is initialized, every tab should have a schema, even if just {}
        schema = self.schemas.get(tab_name)
        if schema is None:
            raise ValueError(f"No schema was given or fetched for tab {tab_name!r}.")
        return schema

    def note_problem(self, problem: str):
        self._problems.append(problem)

    def build_lookup_table_for_tab(self, tab_name: str, *, rows: List[Dict]) -> Dict[str, Dict]:
        if not self.schema_manager or not rows:
            return []
        identifying_properties = self.schema_manager.identifying_properties(schema_name=tab_name)
        if not identifying_properties:
            # Maybe issue a warning here that we're going to lose
            empty_lookup_table: Dict[str, Dict] = {}
            return empty_lookup_table
        lookup_table: Dict[str, Dict] = {}
        for row in rows:
            for identifying_property in identifying_properties:
                value = row.get(identifying_property)
                if value != '' and value is not None:
                    lookup_table[str(value)] = row
        return lookup_table

    def contains_ref(self, item_type, item_ref):
        ref = self.resolve_ref(item_type=item_type, item_ref=item_ref)
        if ref is None:
            return False
        else:
            return True

    def resolve_ref(self, item_type, item_ref):
        lookup_table = self.lookup_tables_by_tab_name.get(item_type)
        if lookup_table:  # Is it a type we're tracking?
            if isinstance(item_ref, list):
                found = True
                for item in item_ref:
                    if not lookup_table.get(item):
                        found = False
                        break
                return True if found else None
            else:
                return lookup_table.get(item_ref) or None
        else:  # Apparently some stray type not in our tables
            return None

    def raise_any_pending_problems(self):
        problems = self._problems
        if problems:
            for problem in problems:
                PRINT(f"Problem: {problem}")
            raise ValidationProblem(problems)
            # raise Exception(there_are(problems, kind='problem while compiling hints', tense='past', show=False))

    def check_tabs(self):
        result = {tab_name: self.check_tab(tab_name)
                  for tab_name in self.tab_names}
        # At this point, doing the checking will have already raised certain errors, if those errors interfere
        # with continued checking, but some smaller problems may have been deferred until the end, so we have to
        # check for and raise an error for any such pending problems now.
        self.raise_any_pending_problems()
        return result

    def validate_ref(self, item_type, item_ref):
        if self.contains_ref(item_type=item_type, item_ref=item_ref):
            return True
        try:
            # TODO: This probably needs a cache
            if isinstance(item_ref, list):
                found = True
                for item in item_ref:
                    info = get_metadata(f"/{to_camel_case(item_type)}/{item}",
                                        ff_env=self.portal_env, vapp=self.portal_vapp)
                    if not isinstance(info, dict) or 'uuid' not in info:
                        found = False
                        break
                return found
            else:
                info = get_metadata(f"/{to_camel_case(item_type)}/{item_ref}",
                                    ff_env=self.portal_env, vapp=self.portal_vapp)
            # Basically return True if there's a value at all,
            # but still check it's not an error message that didn't get raised.
            return isinstance(info, dict) and 'uuid' in info
        except Exception:
            return False

    def schema_exists(self, schema_name: str) -> bool:
        return self.schema_manager.schema_exists(schema_name) if self.schema_manager else False

    def check_tab(self, tab_name: str):
        prototype = self.patch_prototypes_by_tab_name[tab_name]
        parsed_headers = self.parsed_headers_by_tab_name[tab_name]
        type_hints = self.type_hints_by_tab_name[tab_name]
        result = [self.check_row(row, tab_name=tab_name, row_number=row_number, prototype=prototype,
                                 parsed_headers=parsed_headers, type_hints=type_hints)
                  for row_number, row in enumerate(self.tabbed_sheet_data[tab_name])]
        return result

    def check_row(self, row: Dict, *, tab_name: str, row_number: int, prototype: Dict,
                  parsed_headers: ParsedHeaders, type_hints: OptionalTypeHints):
        if self.flattened:
            return self.check_flattened_row(row=row, tab_name=tab_name, row_number=row_number, prototype=prototype,
                                            parsed_headers=parsed_headers, type_hints=type_hints)
        else:
            return self.check_inflated_row(row=row, tab_name=tab_name, row_number=row_number, prototype=prototype,
                                           parsed_headers=parsed_headers, type_hints=type_hints)

    def check_inflated_row(self, row: Dict, *, tab_name: str, row_number: int, prototype: Dict,
                           parsed_headers: ParsedHeaders, type_hints: OptionalTypeHints):
        ignorable(self, tab_name, row_number, prototype, parsed_headers, type_hints)  #
        # TODO: Make this work...
        # def traverse(item, *, subschema, breadcrumbs):
        #     if isinstance(item, list):
        #         # check schema here to make sure it's supposed to be a list before proceeding
        #         for i, elem in enumerate(item):
        #             traverse(item, subschema=..., breadcrumbs=(*breadcrumbs, i))
        #     elif isinstance(item, dict):
        #         # check schema here to make sure it's supposed to be a dict before proceeding
        #         for k, v in item.items():
        #             traverse(v, subschema=..., breadcrumbs=(*breadcrumbs, k))
        #     else:
        #         # look up hint. if there's not a hint for these breadcrumbs, make one
        #         # apply the hint for side-effect, to get an error if we have a bad value
        #         pass
        # schema = self.schemas[tab_name]
        # if schema:
        #     traverse(row, subschema=schema, breadcrumbs=())  # for side-effect
        return row

    def check_flattened_row(self, row: Dict, *, tab_name: str, row_number: int, prototype: Dict,
                            parsed_headers: ParsedHeaders, type_hints: OptionalTypeHints):
        patch_item = copy.deepcopy(prototype)
        for column_number, column_value in enumerate(row.values()):
            parsed_value = ItemTools.parse_item_value(column_value, apply_heuristics=self.apply_heuristics)
            column_name = (list(row.keys())[column_number] or "") if len(row) > column_number else ""
            if column_name.endswith("#"):
                if isinstance(parsed_value, str):
                    parsed_value = [value.strip() for value in parsed_value.split(ARRAY_VALUE_DELIMITER) if value]
            type_hint = type_hints[column_number]
            if type_hint:
                try:
                    src = f"{tab_name}{f'.{column_name}' if column_name else ''}"
                    parsed_value = type_hint.apply_hint(parsed_value, src)
                except ValidationProblem as e:
                    headers = self.headers_by_tab_name[tab_name]
                    column_name = headers[column_number]
                    self.note_problem(f"{tab_name}[{row_number}].{column_name}: {e}")
            ItemTools.set_path_value(patch_item, parsed_headers[column_number], parsed_value)
        return patch_item

    @classmethod
    def check(cls, tabbed_sheet_data: TabbedSheetData, *,
              flattened: bool,
              override_schemas: Optional[TabbedJsonSchemas] = None,
              apply_heuristics: bool = False,
              portal_env: Optional[str] = None, portal_vapp: Optional[AbstractVirtualApp] = None):
        checker = cls(tabbed_sheet_data, flattened=flattened,
                      override_schemas=override_schemas, apply_heuristics=apply_heuristics,
                      portal_env=portal_env, portal_vapp=portal_vapp)
        checked = checker.check_tabs()
        return checked

    class SheetState:

        def __init__(self, parsed_headers: ParsedHeaders, type_hints: OptionalTypeHints):
            self.parsed_headers = parsed_headers
            self.type_hints = type_hints

    def compile_type_hints(self, tab_name: str) -> OptionalTypeHints:
        parsed_headers = self.parsed_headers_by_tab_name[tab_name]
        schema = self.schemas.get(tab_name)
        for required_header in self._schema_required_headers(schema):
            if required_header not in parsed_headers:
                self.note_problem("Missing required header")
        positional_type_hints = [(ItemTools.find_type_hint_for_parsed_header(parsed_header, schema, context=self)
                                  if schema
                                  else None)
                                 for parsed_header in parsed_headers]
        type_hints = OptionalTypeHints(positional_type_hints, positional_breadcrumbs=parsed_headers)
        return type_hints

    @classmethod
    def _schema_required_headers(cls, schema):
        ignored(schema)
        return []  # TODO: Make this compute a list of required headers (in parsed header form)

    def create_tab_processor_state(self, tab_name: str) -> SheetState:
        # This will create state that allows us to efficiently assign values in the right place on each row
        return self.SheetState(parsed_headers=self.parsed_headers_by_tab_name[tab_name],
                               type_hints=self.type_hints_by_tab_name[tab_name])


# check = TableChecker.check


def load_items(filename: str, tab_name: Optional[str] = None, escaping: Optional[bool] = None,
               override_schemas: Optional[TabbedJsonSchemas] = None, apply_heuristics: bool = False,
               portal_env: Optional[str] = None, portal_vapp: Optional[AbstractVirtualApp] = None,
               # TODO: validate= is presently False (i.e., disabled) by default while being debugged,
               #       but for production use maybe should not be? -kmp 25-Oct-2023
               validate: bool = False,
               retain_empty_properties: bool = False,
               sheet_order: Optional[List[str]] = None,
               **kwargs):
    annotated_data = TableSetManager.load_annotated(filename=filename, tab_name=tab_name, escaping=escaping,
                                                    prefer_number=False, sheet_order=sheet_order, **kwargs)
    tabbed_rows = annotated_data['content']
    flattened = annotated_data['flattened']
    if flattened:
        checked_items = TableChecker.check(tabbed_rows, flattened=flattened,
                                           override_schemas=override_schemas,
                                           portal_env=portal_env, portal_vapp=portal_vapp,
                                           apply_heuristics=apply_heuristics)
    else:
        # No fancy checking for things like .json, etc. for now. Only check things that came from
        # spreadsheet-like data, where structural datatypes are forced into strings.
        checked_items = tabbed_rows
    if not retain_empty_properties:
        remove_empty_properties(checked_items)
    if validate:
        problems = validate_data_against_schemas(checked_items, portal_env=portal_env, portal_vapp=portal_vapp,
                                                 override_schemas=override_schemas)
        return checked_items, problems
    return checked_items



from contextlib import contextmanager
from datetime import datetime
import os
import shutil
import tempfile
from uuid import uuid4 as uuid
from typing import List, Optional, Union
from dcicutils.file_utils import create_random_file


@contextmanager
def temporary_directory() -> str:
    try:
        with tempfile.TemporaryDirectory() as tmp_directory_name:
            yield tmp_directory_name
    finally:
        remove_temporary_directory(tmp_directory_name)


@contextmanager
def temporary_file(name: Optional[str] = None, prefix: Optional[str] = None, suffix: Optional[str] = None,
                   content: Optional[Union[str, bytes, List[str]]] = None) -> str:
    with temporary_directory() as tmp_directory_name:
        tmp_file_name = f"{prefix or ''}{name or tempfile.mktemp(dir='')}{suffix or ''}"
        tmp_file_path = os.path.join(tmp_directory_name, tmp_file_name)
        with open(tmp_file_path, "wb" if isinstance(content, bytes) else "w") as tmp_file:
            if content is not None:
                tmp_file.write("\n".join(content) if isinstance(content, list) else content)
        yield tmp_file_path


def create_temporary_file_name(prefix: Optional[str] = None, suffix: Optional[str] = None) -> str:
    """
    Generates and returns the full path to file within the system temporary directory.
    """
    random_string = f"{datetime.utcnow().strftime('%Y%m%d%H%M%S')}{str(uuid()).replace('-', '')}"
    tmp_file_name = f"{prefix or ''}{random_string}{suffix or ''}"
    return os.path.join(tempfile.gettempdir(), tmp_file_name)


@contextmanager
def temporary_random_file(prefix: Optional[str] = None, suffix: Optional[str] = None,
                          nbytes: int = 1024, binary: bool = False, line_length: Optional[int] = None) -> str:
    with temporary_file(prefix=prefix, suffix=suffix) as tmp_file_path:
        create_random_file(tmp_file_path, nbytes=nbytes, binary=binary, line_length=line_length)
        yield tmp_file_path


def remove_temporary_directory(tmp_directory_name: str) -> None:
    """
    Removes the given directory, recursively; but ONLY if it is (somewhere) within the system temporary directory.
    """
    if is_temporary_directory(tmp_directory_name):  # Guard against errant deletion.
        shutil.rmtree(tmp_directory_name)


def remove_temporary_file(tmp_file_name: str) -> bool:
    """
    Removes the given file; but ONLY if it is (somewhere) within the system temporary directory.
    """
    try:
        tmpdir = tempfile.gettempdir()
        if (os.path.commonpath([tmpdir, tmp_file_name]) == tmpdir) and os.path.isfile(tmp_file_name):
            os.remove(tmp_file_name)
            return True
        return False
    except Exception:
        return False


def is_temporary_directory(path: str) -> bool:
    try:
        tmpdir = tempfile.gettempdir()
        return os.path.commonpath([path, tmpdir]) == tmpdir and os.path.exists(path) and os.path.isdir(path)
    except Exception:
        return False



import argparse
import botocore.exceptions
import boto3
import contextlib
import os
import sys

from typing import Optional, Union, List
from .command_utils import yes_or_no
from .ecr_utils import ECRUtils
from .misc_utils import PRINT, full_class_name
from .lang_utils import n_of


EPILOG = __doc__


DEFAULT_IMAGE_REPOSITORY = ECRUtils.DEFAULT_IMAGE_REPOSITORY
DEFAULT_ECS_REPOSITORY = DEFAULT_IMAGE_REPOSITORY  # support for deprecated name

# This is the currently selected environment, though commands might want to require confirmation before using it.
DEFAULT_ACCOUNT_NUMBER = os.environ.get('ACCOUNT_NUMBER')

# Use if an unsupplied value requires interactive confirmation before defaulting
CONFIRM_DEFAULT_ACCOUNT_NUMBER = 'confirm-default-account-number'

IMAGE_COUNT_LIMIT = ECRUtils.IMAGE_LIST_DEFAULT_COUNT_LIMIT  # Deprecated (to remove in release 5.0) -kmp 4-Sep-2022
IMAGE_LIST_CHUNK_SIZE = ECRUtils.IMAGE_LIST_CHUNK_SIZE  # Deprecated (to remove in release 5.0) -kmp 4-Sep-2022
RELEASED_TAG = ECRUtils.IMAGE_RELEASED_TAG  # Deprecated (to remove in release 5.0) -kmp 4-Sep-2022


def make_sorted_string_list(elements, separator=","):
    """
    Args:
        elements: a sequence such as a list that can be sorted
        separator: a separator to use between items when the collection is more than a single element.
    Returns:
        a comma-separated string
    """
    return separator.join(sorted(elements))


@contextlib.contextmanager
def ecr_command_context(account_number, ecs_repository=None, ecr_client=None):
    """
    A context manager that takes an account AWS account number specified by the user in a command, and does
    plausibility checking on that account number against the value of the environment variable ACCOUNT_NUMBER.
    If no error occurs, it proceeds to execute a body of code in a context that traps errors in a manner
    appropriate to a script invoked from a command line.

    Args:
        account_number: an AWS account number (represented as a string)
        ecs_repository: The name of an ecs_repository, such as 'main'.
        ecr_client: an AWS ecr_client to use instead of having to make one.
    """
    try:
        try:
            if account_number == CONFIRM_DEFAULT_ACCOUNT_NUMBER:
                if yes_or_no(f"Use {DEFAULT_ACCOUNT_NUMBER} as account number?"):
                    account_number = DEFAULT_ACCOUNT_NUMBER
                else:
                    account_number = None
            if not account_number:
                raise RuntimeError("You must specify an AWS account number.")
            account_number_in_environ = os.environ.get('ACCOUNT_NUMBER')
            if not account_number_in_environ:
                raise RuntimeError("You do not have globally declared credentials."
                                   " As part of setting them, ACCOUNT_NUMBER would be set.")
            elif account_number != account_number_in_environ:
                raise RuntimeError("The account number you have specified does not match your declared credentials.")
            yield ECRCommandContext(account_number=account_number, ecs_repository=ecs_repository, ecr_client=ecr_client)
            sys.exit(0)
        except botocore.exceptions.ClientError as e:
            error_info = e.response.get('Error', {})
            message = error_info.get('Message')
            code = error_info.get('Code')
            if code == 'ExpiredTokenException':
                raise RuntimeError("Your security token seems to have expired.")
            elif message:
                PRINT(f"{code}: {message}")
                sys.exit(1)
            else:
                raise
    except Exception as e:
        PRINT(f"{full_class_name(e)}: {e}")
        sys.exit(1)


class ECRCommandContext:

    def __init__(self, account_number=None, ecs_repository=None, image_repository=None, ecr_client=None):
        """
        Initializes an ECRContextText

        Args:
            account_number: an AWS account number (represented as a string)
            ecs_repository: The name of an ecs_repository, such as 'main'.
            ecr_client: an AWS ecr_client to use instead of having to make one.
        """
        self.account_number = account_number or DEFAULT_ACCOUNT_NUMBER
        self.image_repository = (image_repository
                                 or ecs_repository  # Deprecated
                                 or DEFAULT_IMAGE_REPOSITORY
                                 or DEFAULT_ECS_REPOSITORY)  # Deprecated
        self.ecr_client = ecr_client or boto3.client('ecr', region_name=ECRUtils.REGION)
        self.ecr_utils = ECRUtils(image_repository=self.image_repository, ecr_client=self.ecr_client)

    def get_images_descriptions(self,
                                digests: Optional[List[str]] = None,
                                tags: Optional[List[str]] = None,
                                limit: Optional[Union[int, str]] = ECRUtils.IMAGE_LIST_DEFAULT_COUNT_LIMIT):
        """
        Args:
            digests: a list of image digests (each represented as a string in sha256 format)
            tags: a list of strings (each being a potential or actual image tag).
            limit: an int indicating a limit on how many results to return,
                   a string indicating a limiting tag,
                   or None indicating no limit.

        Returns:
            a dictionary with keys 'descriptions' (the primary result), 'count' (the number of descriptions),
            and 'total' (the total number of registered images in the catalog).
            The descriptions are a list of individual dictionaries with keys that include, at least,
            'imagePushedAt', 'imageTags', and 'imageDigest'.
        """
        return self.ecr_utils.get_images_descriptions(digests=digests, tags=tags, limit=limit)

    def show_image_catalog(self,
                           digests: Optional[List[str]] = None,
                           tags: Optional[List[str]] = None,
                           limit: Optional[Union[int, str]] = ECRUtils.IMAGE_LIST_DEFAULT_COUNT_LIMIT,
                           summarize: bool = True):
        """
        Shows a list of tabular information about images in the given account number and ECS respository.

        Args:
            digests: a list of image digests (each represented as a string in sha256 format)
            tags: a list of strings (each being a potential or actual image tag).
            limit: an int indicating a limit on how many results to return,
                   a string indicating a limiting tag,
                   or None indicating no limit.
            summarize: a boolean saying whether to show a summary at the end
        """
        info = self.get_images_descriptions(digests=digests, tags=tags, limit=limit)
        for image_detail in info['descriptions']:
            image_pushed_at = image_detail['imagePushedAt']
            image_tags = image_detail.get('imageTags', [])
            image_digest = image_detail['imageDigest']
            PRINT(f"{image_digest} {image_pushed_at} {make_sorted_string_list(image_tags, separator=', ')}")
        n = info['count']
        total = info['total']
        if summarize:
            current_account = f"account {self.account_number}" if self.account_number else "current account"
            if digests or tags:
                PRINT(f"Only {n_of(n, 'relevant image')} in {current_account} shown.")
            else:
                PRINT(f"{'All' if n == total else 'Most recent'} {info['count']} of {info['total']} images"
                      f" in {current_account} shown.")

    def get_image_manifest(self, tag=None, digest=None) -> str:
        """
        Returns the imageManifest of an image given a tag and/or digest describing that image.
        (There must be only one matching image.)

        Args:
            digest: a list of image digests (each represented as a string in sha256 format)
            tag: a list of strings (each being a potential or actual image tag).
        Returns:
            a string which is the imageManifest
        Raises:
            RuntimeError: if no images match, or if too many images match
        """
        if tag and not digest:
            image_id = {"imageTag": tag}
        elif digest and not tag:
            image_id = {"imageDigest": digest}
        else:  # neither or both
            raise RuntimeError("Exactly one of --tag and --digest is required.")
        options = {'repositoryName': self.image_repository, 'imageIds': [image_id]}
        response = self.ecr_client.batch_get_image(**options)
        images = response.get('images')
        if not images or len(images) == 0:
            raise RuntimeError(f"No match for {image_id}.")
        elif len(images) > 1 and not all(image['imageManifest'] == images[0]['imageManifest'] for image in images):
            # Usually we get more than one because it's the same image with different labels,
            # but this is just in case the problem is other than that.
            raise RuntimeError(f"Too many ({len(images)}) matches for {image_id}.")
        else:
            image = images[0]
        return image.get('imageManifest')

    def show_image_manifest(self, digest=None, tag=None):
        """
        Shows the 'imageManifest' for a given deployment image, which is the string representation of a JSON dictionary,
        as if json.dumps() of that dictionary had been printed.

        NOTE: This command may be useful primarily for debugging. This manifest is needed to support add_image_tag,
              but most people don't care about seeing this data.

        Args:
            digest: a list of image digests (each represented as a string in sha256 format)
            tag: a list of strings (each being a potential or actual image tag).
        """

        info = self.get_image_manifest(tag=tag, digest=digest)
        PRINT(info)

    def add_image_tag(self, digest, tag):
        """
            digest: a list of image digests (each represented as a string in sha256 format)
            tag: a list of strings (each being a potential or actual image tag).
        """
        relevant_digests = [digest]
        try:
            # This figures out where the image was before. It might be it was nowhere, in which case nothing gets added.
            info = self.get_images_descriptions(tags=[tag])
            old_digest = info['descriptions'][0]['imageDigest']
            relevant_digests.append(old_digest)
        except Exception:
            pass
        old_tags = []
        try:
            # This figures out where the image was before. It might be it was nowhere, in which case nothing gets added.
            info = self.get_images_descriptions(digests=[digest])
            old_tags = info['descriptions'][0]['imageTags']
        except Exception:
            pass
        if tag in old_tags:
            raise RuntimeError(f"Image {digest} is already labeled {tag!r}.")
        if tag.startswith("sha256:"):
            raise RuntimeError(f"Invalid tag: {tag}")
        if not digest.startswith("sha256:"):
            raise RuntimeError(f"Invalid digest (not a sha256): {digest}")
        PRINT("Before:")
        self.show_image_catalog(digests=relevant_digests, summarize=False)
        manifest = self.get_image_manifest(digest=digest)
        self.ecr_client.put_image(repositoryName=self.image_repository, imageManifest=manifest, imageTag=tag)
        # This shows what changed.
        PRINT("After:")
        self.show_image_catalog(digests=relevant_digests)

    def _most_recent_two_deploys(self):
        """
        Returns:
            the most recent two of several deploys
        Raises:
            RuntimeError: if there are not at least two
        """
        info = self.get_images_descriptions()
        descriptions = info.get('descriptions')
        if not descriptions:
            raise RuntimeError("Nothing seems to have been deployed.")
        elif len(descriptions) == 1:
            raise RuntimeError("There is only one deploy.")
        else:
            most_recent_two = descriptions[:2]
            return most_recent_two

    def unrelease_most_recent_image(self):
        """
        Moves the released tag from the most recent deploy image to the penultimate deploy image.
        This sets up for a Fargate redeploy to find that previous build when it next redeploys ECS.
        but it does not actually invoke that redeploy.
        """
        [most_recent, previous] = self._most_recent_two_deploys()
        if RELEASED_TAG not in most_recent.get('imageTags', []):
            self.show_image_catalog(limit=RELEASED_TAG)
            raise RuntimeError(f"Most recent deploy is NOT marked {RELEASED_TAG!r}."
                               f" You'll have to sort this out manually.")
        self.add_image_tag(digest=previous['imageDigest'], tag=RELEASED_TAG)


def show_image_catalog_main(override_args=None):
    """
    This is the main program that can be used from the command line to invoke the show_image_catalog function.

    Args:
        override_args: a list of arguments to  use instead of the command line arguments (usually for testing)
    """

    parser = argparse.ArgumentParser(  # noqa - PyCharm wrongly thinks the formatter_class is invalid
        description="Show descriptions of all images in ECS", epilog=EPILOG,
        formatter_class=argparse.RawDescriptionHelpFormatter,
    )
    parser.add_argument('--account-number', dest='account_number',
                        default=DEFAULT_ACCOUNT_NUMBER, metavar="<acct-num>",
                        help=f"AWS account number containing repository (default: {DEFAULT_ACCOUNT_NUMBER})")
    parser.add_argument('--ecs-repository', dest='ecs_repository',
                        default=DEFAULT_IMAGE_REPOSITORY, metavar="<repo-name>",
                        help=f"repository name to show images for (default: {DEFAULT_IMAGE_REPOSITORY})")
    parser.add_argument('--limit', default=ECRUtils.IMAGE_LIST_DEFAULT_COUNT_LIMIT, metavar="<n>", type=int,
                        help=f"maximum number of images to describe"
                             f" (default: {ECRUtils.IMAGE_LIST_DEFAULT_COUNT_LIMIT})")
    parsed = parser.parse_args(args=override_args)

    with ecr_command_context(account_number=parsed.account_number,
                             ecs_repository=parsed.ecs_repository) as command_context:
        command_context.show_image_catalog(limit=parsed.limit)


def show_image_manifest_main(override_args=None):
    """
    This is the main program that can be used from the command line to invoke the show_image_manifest function.

    NOTE: This command may be useful primarily for debugging. This manifest is needed to support add_image_tag,
          but most people don't care about seeing this data.

    Args:
        override_args: a list of arguments to  use instead of the command line arguments (usually for testing)
    """

    parser = argparse.ArgumentParser(  # noqa - PyCharm wrongly thinks the formatter_class is invalid
        description="Show the manifest for an ECS image, specified by either label or digest (sha).", epilog=EPILOG,
        formatter_class=argparse.RawDescriptionHelpFormatter,
    )
    parser.add_argument('--account-number', dest='account_number',
                        default=DEFAULT_ACCOUNT_NUMBER, metavar="<acct-num>",
                        help=f"AWS account number containing repository (default: {DEFAULT_ACCOUNT_NUMBER})")
    parser.add_argument('--ecs-repository', dest='ecs_repository',
                        default=DEFAULT_IMAGE_REPOSITORY, metavar="<repo-name>",
                        help=f"repository name to show images for (default: {DEFAULT_IMAGE_REPOSITORY})")
    parser.add_argument('--digest', default=None, metavar="<sha>",
                        help=f"digest (sha) to look for (required unless --tag is given)")
    parser.add_argument('--tag', default=None, metavar="<tag>",
                        help=f"tag to look for (required unless --digest is given)")
    parsed = parser.parse_args(args=override_args)

    with ecr_command_context(account_number=parsed.account_number,
                             ecs_repository=parsed.ecs_repository) as command_context:
        command_context.show_image_manifest(tag=parsed.tag, digest=parsed.digest)


def add_image_tag_main(override_args=None):
    """
    This is the main program that can be used from the command line to invoke the add_image_tag function.

    Args:
        override_args: a list of arguments to  use instead of the command line arguments (usually for testing)
    """
    parser = argparse.ArgumentParser(  # noqa - PyCharm wrongly thinks the formatter_class is invalid
        description="Adds a tag to a given image.", epilog=EPILOG,
        formatter_class=argparse.RawDescriptionHelpFormatter,
    )
    parser.add_argument('--account-number', dest='account_number',
                        default=CONFIRM_DEFAULT_ACCOUNT_NUMBER, metavar="<acct-num>",
                        help=f"AWS account number containing repository"
                             f" (default: {DEFAULT_ACCOUNT_NUMBER}, requires interactive confirmation if unsupplied)")
    parser.add_argument('--ecs-repository', dest='ecs_repository',
                        default=DEFAULT_IMAGE_REPOSITORY, metavar="<repo-name>",
                        help=f"repository name to show images for (default: {DEFAULT_IMAGE_REPOSITORY})")
    parser.add_argument('digest', metavar="<sha>",
                        help="digest (sha) of image to add tag to (REQUIRED)")
    parser.add_argument('tag', metavar="<tag>",
                        help="tag to add (REQUIRED)")
    parsed = parser.parse_args(args=override_args)

    with ecr_command_context(account_number=parsed.account_number,
                             ecs_repository=parsed.ecs_repository) as command_context:
        command_context.add_image_tag(tag=parsed.tag, digest=parsed.digest)


def unrelease_most_recent_image_main(override_args=None):
    """
    This is the main program that can be used from the command line to invoke the unrelease_most_recent_image function.

    Args:
        override_args: a list of arguments to  use instead of the command line arguments (usually for testing)
    """
    parser = argparse.ArgumentParser(  # noqa - PyCharm wrongly thinks the formatter_class is invalid
        description=f"Moves the {RELEASED_TAG!r} tag from most recent deploy to the previous deploy.", epilog=EPILOG,
        formatter_class=argparse.RawDescriptionHelpFormatter,
    )
    parser.add_argument('--account-number', dest='account_number',
                        default=CONFIRM_DEFAULT_ACCOUNT_NUMBER, metavar="<acct-num>",
                        help=f"AWS account number containing repository"
                             f" (default: {DEFAULT_ACCOUNT_NUMBER}, requires interactive confirmation if unsupplied)")
    parser.add_argument('--ecs-repository', dest='ecs_repository',
                        default=DEFAULT_IMAGE_REPOSITORY, metavar="<repo-name>",
                        help=f"repository name to show images for (default: {DEFAULT_IMAGE_REPOSITORY})")
    parsed = parser.parse_args(args=override_args)

    with ecr_command_context(account_number=parsed.account_number,
                             ecs_repository=parsed.ecs_repository) as command_context:
        command_context.unrelease_most_recent_image()



import secrets
import datetime
import structlog
import jwt
from dcicutils.redis_utils import RedisBase, RedisException


log = structlog.getLogger(__name__)


SESSION_TOKEN_COOKIE = 'c4_st'


def make_session_token(n_bytes: int = 32) -> str:
    """ Uses the secrets module to create a cryptographically secure and URL safe string
    :param n_bytes: number of bytes to use, default 32
    :return: cryptographically secure url safe token for use as a session token
    """
    return secrets.token_urlsafe(n_bytes)


class RedisSessionToken:
    """
    Model used by Redis to store session tokens
    Keystore structure:
        <env_namespace>:session:<token>
            -> Redis str of JWT with automated expiration
    """

    @staticmethod
    def _build_session_expiration(n_hours: int = 3) -> datetime.timedelta:
        """ Builds a session expiration date 3 hours after generation
        :param n_hours: timestamp after which the session token is invalid
        :return: datetime string 3 hours from creation time
        """
        return datetime.timedelta(hours=n_hours)

    @staticmethod
    def _build_redis_key(namespace: str, token: str) -> str:
        """ Builds the hash key used by Redis
        :param namespace: namespace to build keys under, for example the env name
        :param token: value of the token
        :return: redis hash key to store values under
        """
        return f'{namespace}:session:{token}'

    def __init__(self, *, namespace: str, jwt: str, email: str, token=None, expiration=None):
        """ Creates a Redis Session object, storing a hash of the JWT into Redis and returning this
            value as the session token.
            :param namespace: namespace to build key under, for example the env name
            :param jwt: jwt generated for this user
            :param email: email verified for this user
            :param token: value of token if passed, if not one will be generated
            :param expiration: expiration of token if passed, if not new expiration will be generated
        """
        if token:
            self.session_token = token
        else:
            self.session_token = make_session_token()
        self.namespace = namespace
        self.redis_key = self._build_redis_key(self.namespace, self.session_token)
        self.jwt = jwt
        self.email = email
        self.expiration = expiration or self._build_session_expiration()

    def __eq__(self, other):
        """ Evaluates equality of two session objects based on the value of the session hset """
        return (self.redis_key == other.redis_key) and (self.jwt == other.jwt) and (self.email == other.email)

    def get_session_token(self) -> str:
        """ Extracts the session token stored on this object """
        return self.session_token

    def get_redis_key(self) -> str:
        """ Returns the key under which the Redis set is stored - note that this contains the token! """
        return self.redis_key

    def get_expiration(self) -> datetime:
        """ Returns the expiration timedelta """
        return self.expiration

    def get_jwt(self) -> str:
        """ Returns the JWT set on this session token object """
        return self.jwt

    def get_email(self) -> str:
        """ Returns the email set on this session token object """
        return self.email

    @classmethod
    def from_redis(cls, *, redis_handler: RedisBase, namespace: str, token: str):
        """ Builds a RedisSessionToken from an existing record - allows extracting JWT
            given a session token internally.
        :param redis_handler: handle to Redis API
        :param namespace: namespace to search under
        :param token: value of the token
        :return: A RedisSessionToken object built from an existing record in Redis
        """
        redis_key = f'{namespace}:session:{token}'
        redis_entry = redis_handler.get(redis_key)
        if redis_entry:
            jwt_and_email = redis_entry.split(':')
            jwt = jwt_and_email[0]
            email = jwt_and_email[1] if len(jwt_and_email) > 1 else None
            expiration = redis_handler.ttl(redis_key)
            return cls(namespace=namespace, jwt=jwt, email=email,
                       token=token, expiration=expiration)

    def decode_jwt(self, audience: str, secret: str, leeway: int = 30, algorithms: list = ['HS256']) -> dict:
        """ Decodes JWT to grab info such as the email
        :param audience: audience under which to decode, typically Auth0Client
        :param secret: secret to decrypt using, typically Auth0Secret
        :param leeway: numerical value in seconds to account for clock drift
        :return: a decoded JWT in dictionary format
        """
        return jwt.decode(self.jwt, secret, audience=audience, leeway=leeway,
                          options={'verify_signature': True}, algorithms=algorithms)

    def store_session_token(self, *, redis_handler: RedisBase) -> bool:
        """ Stores the created session token object as an hset in Redis
        :param redis_handler: handle to Redis API
        :return: True if successful, raise Exception otherwise
        """
        try:
            redis_handler.set(self.redis_key, f'{self.jwt}:{self.email or ""}', exp=self.expiration)
        except Exception as e:
            log.error(str(e))
            raise RedisException()
        return True

    def validate_session_token(self, *, redis_handler: RedisBase) -> bool:
        """ Validates the given session token against that stored in redis
        :param redis_handler: handle to Redis API
        :return: True if token matches that in Redis and is not expired
        """
        redis_token = redis_handler.get(self.redis_key)
        if not redis_token:
            return False  # if it doesn't exist it's not valid
        return True  # if it does exist it must be valid since we always send with TTL

    def update_session_token(self, *, redis_handler: RedisBase, jwt: str, email: str) -> bool:
        """ Refreshes the session token, jwt (if different) and expiration stored in Redis
        :param redis_handler: handle to Redis API
        :param jwt: jwt of user
        :param email: email of user
        :return: True if successful, raise Exception otherwise
        """
        # remove old token
        self.delete_session_token(redis_handler=redis_handler)
        # build new one
        self.session_token = make_session_token()
        self.redis_key = self._build_redis_key(self.namespace, self.session_token)
        self.expiration = self._build_session_expiration()
        self.jwt = jwt
        self.email = email
        return self.store_session_token(redis_handler=redis_handler)

    def delete_session_token(self, *, redis_handler: RedisBase) -> bool:
        """ Deletes the session token from redis, effectively logging out
        :param redis_handler: handle to Redis API
        :return: True if successful, False otherwise
        """
        return 1 == redis_handler.delete(self.redis_key)
